'''dossier.label.tests

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

from pyquchk import qc

from dossier.label import CorefValue, Label
from dossier.label.tests import coref_value, id_, time_value


@qc
def test_label_reverse_equality(cid1=id_, cid2=id_, ann=id_, v=coref_value,
                                t=time_value):
    l1 = Label(cid1, cid2, ann, v, epoch_ticks=t)
    l2 = Label(cid2, cid1, ann, v, epoch_ticks=t)
    assert l1 == l2
    assert hash(l1) == hash(l2)


@qc
def test_content_id_order(cid1=id_, cid2=id_, ann=id_, v=coref_value):
    l = Label(cid1, cid2, ann, v)
    assert cid1 in l
    assert cid2 in l
    assert l.content_id1 <= l.content_id2
    assert l.content_id1 == min(cid1, cid2)
    assert l.content_id2 == max(cid1, cid2)


@qc
def test_subtopic_order(cid=id_, s1=id_, s2=id_, ann=id_, v=coref_value):
    l = Label(cid, cid, ann, v, subtopic_id1=s1, subtopic_id2=s2)
    assert cid in l
    assert (cid, s1) in l
    assert (cid, s2) in l
    assert l.content_id1 == cid
    assert l.content_id2 == cid
    assert l.subtopic_id1 <= l.subtopic_id2
    assert l.subtopic_id1 == min(s1, s2)
    assert l.subtopic_id2 == max(s1, s2)


@qc
def test_subtopic_id(cid1=id_, cid2=id_, s1=id_, s2=id_, ann=id_,
                     v=coref_value):
    l = Label(cid1, cid2, ann, v, subtopic_id1=s1, subtopic_id2=s2)
    assert cid1 in l
    assert (cid1, None) in l
    assert (cid1, s1) in l
    assert cid2 in l
    assert (cid2, None) in l
    assert (cid2, s2) in l
    assert l.other(cid1) == cid2
    assert l.other(cid2) == cid1
    if cid1 != cid2:
        assert l.subtopic_for(cid1) == s1
        assert l.subtopic_for(cid2) == s2
    else:
        assert l.subtopic_for(cid1) == min(s1, s2)


@qc
def test_same_subject(cid1=id_, cid2=id_, s1=id_, s2=id_, ann=id_,
                      v1=coref_value, v2=coref_value,
                      t1=time_value, t2=time_value):
    l1 = Label(cid1, cid2, ann, v1, epoch_ticks=t1,
               subtopic_id1=s1, subtopic_id2=s2)
    l2 = Label(cid1, cid2, ann, v2, epoch_ticks=t2,
               subtopic_id1=s1, subtopic_id2=s2)
    assert l1.same_subject_as(l2)
    assert l2.same_subject_as(l1)


@qc
def test_label_most_recent_first(cid1=id_, cid2=id_, ann=id_,
                                 v1=coref_value, v2=coref_value,
                                 t=time_value):
    lab1 = Label(cid1, cid2, ann, v1, epoch_ticks=t)
    lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=t + 1)
    assert lab2 < lab1
    assert not (lab1 == lab2)
    assert sorted([lab1, lab2]) == [lab2, lab1]
    assert list(Label.most_recent([lab2, lab1])) == [lab2]


@qc
def test_label_most_recent_first_unordered(cid1=id_, cid2=id_, ann=id_,
                                           v1=coref_value, v2=coref_value,
                                           t=time_value):
    lab1 = Label(cid1, cid2, ann, v1, epoch_ticks=t)
    lab2 = Label(cid2, cid1, ann, v2, epoch_ticks=t + 1)
    assert lab2 < lab1
    assert not (lab1 == lab2)
    assert sorted([lab1, lab2]) == [lab2, lab1]
    assert list(Label.most_recent([lab2, lab1])) == [lab2]


@qc
def test_label_order_on_value(cid1=id_, cid2=id_, ann=id_, t=time_value,
                              v1=coref_value, v2=coref_value):
    lab1 = Label(cid1, cid2, ann, v1, epoch_ticks=t)
    lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=t)
    assert ((v1 < v2 and lab1 < lab2) or
            (v1 == v2 and lab1 == lab2) or
            (v1 > v2 and lab1 > lab2))


def label(id1, id2, v=CorefValue.Positive, sid1=None, sid2=None):
    return Label(id1, id2, 'foo', v, subtopic_id1=sid1, subtopic_id2=sid2)


def diff_labels_sets(old, new):
    diff = Label.diff(old, new)
    return {
        'add': set(diff['add']),
        'delete': set(diff['delete']),
        'change': set(diff['change']),
    }

def test_label_diff_simple():
    old = [label('a', 'b'), label('a', 'c')]
    new = [label('a', 'b'), label('a', 'd')]
    assert diff_labels_sets(old, new) == {
        'add': {label('a', 'd')},
        'delete': {label('a', 'c')},
        'change': set(),
    }


def test_label_diff_change():
    old = [label('a', 'b')]
    new = [label('a', 'b', v=CorefValue.Negative)]
    assert diff_labels_sets(old, new) == {
        'add': set(),
        'delete': set(),
        'change': {label('a', 'b', v=CorefValue.Negative)},
    }


def test_label_diff_empty():
    old = [Label('a', 'b', 'foo', 1, epoch_ticks=0)]
    new = [Label('a', 'b', 'foo', 1, epoch_ticks=1)]
    assert old != new
    assert diff_labels_sets(old, new) == {
        'add': set(),
        'delete': set(),
        'change': set(),
    }


def test_label_diff_simple_change():
    old = [label('a', 'b'), label('a', 'c')]
    new = [label('a', 'b', v=CorefValue.Negative), label('a', 'd')]
    assert diff_labels_sets(old, new) == {
        'add': {label('a', 'd')},
        'delete': {label('a', 'c')},
        'change': {label('a', 'b', v=CorefValue.Negative)},
    }



'''dossier.label.label

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

import cbor
from collections import Container, Hashable
from datetime import datetime
from itertools import combinations, ifilter, imap, product
import logging
from operator import attrgetter, itemgetter
import struct
import time

import enum
from total_ordering import total_ordering


logger = logging.getLogger(__name__)


MAX_SECOND_TICKS = ((60 * 60 * 24) * 365 * 100)
'''The maximum number of seconds supported.

Our kvlayer backend cannot (currently) guarantee a correct ordering
of signed integers, but can guarantee a correct ordering of unsigned
integers.

Labels, however, should be sorted with the most recent label first.
This is trivially possible by negating its epoch ticks.

Because kvlayer cannot guarantee a correct ordering of signed integers,
we avoid the sign switch by subtracting the ticks from an arbitrary
date in the future (UNIX epoch + 100 years).
'''


def time_complement(t):
    return long(MAX_SECOND_TICKS - t)


class CorefValue(enum.Enum):
    '''A human-assigned value for a coreference judgement.

    The judgment is always made with respect to a pair of content
    items.

    :cvar Negative: The two items are not coreferent.
    :cvar Unknown: It is unknown whether the two items are coreferent.
    :cvar Positive: The two items are coreferent.

    '''
    Negative = -1
    Unknown = 0
    Positive = 1

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented


@total_ordering
class Label(Container, Hashable):
    '''An immutable unit of ground truth data.

    This is a statement that the item at :attr:`content_id1`,
    :attr:`subtopic_id1` refers to (or doesn't) the same thing as the
    item at :attr:`content_id2`, :attr:`subtopic_id2`.  This assertion
    was recorded by :attr:`annotator_id`, a string identifying a user,
    at :attr:`epoch_ticks`, with :class:`CorefValue` :attr:`value`.

    On creation, the tuple is normalized such that the pair of
    `content_id1` and `subtopic_id1` are less than the pair of
    `content_id2` and `subtopic_id2`.

    Labels are comparable, sortable, and hashable.  The sort order
    compares the two content IDs, the two subtopic IDs,
    :attr:`annotator_id`, :attr:`epoch_ticks` (most recent is
    smallest), and then other fields.

    .. attribute:: content_id1

       The first content ID.

    .. attribute:: content_id2

       The second content ID.

    .. attribute:: annotator_id

       An identifier of the user making this assertion.

    .. attribute:: value

       A :class:`CorefValue` stating whether this is a positive or
       negative coreference assertion.

    .. attribute:: subtopic_id1

       An identifier defining a section or region of
       :attr:`content_id1`.

    .. attribute:: subtopic_id2

       An identifier defining a section or region of
       :attr:`content_id2`.

    .. attribute:: epoch_ticks

       The time at which :attr:`annotator_id` made this assertion,
       in seconds since the Unix epoch.

    .. attribute:: rating

       An additional score showing the relative importance of this
       label, mirroring :class:`streamcorpus.Rating`.

    .. attribute:: meta

       Any additional meta data about this label, structured as
       a dictionary.

    .. automethod:: __init__
    .. automethod:: __contains__
    .. automethod:: other
    .. automethod:: subtopic_for
    .. automethod:: same_subject_as
    .. automethod:: most_recent

    '''

    def __init__(self, content_id1, content_id2, annotator_id, value,
                 subtopic_id1=None, subtopic_id2=None, epoch_ticks=None,
                 rating=None, meta=None):
        '''Create a new label.

        Parameters are assigned to their corresponding fields, with
        some normalization.  If `value` is an :class:`int`, the
        corresponding :class:`CorefValue` is used instead.  If
        `epoch_ticks` is :const:`None` then the current time is
        used.  If `rating` is :const:`None` then it will be 1 if
        `value` is :attr:`CorefValue.Positive` and 0 otherwise.

        '''
        super(Label, self).__init__()

        # Fix up and provide defaults for various parameters
        if isinstance(value, int):
            value = CorefValue(value)
        if epoch_ticks is None:
            epoch_ticks = long(time.time())
        if subtopic_id1 is None:
            subtopic_id1 = ''
        if subtopic_id2 is None:
            subtopic_id2 = ''

        if value == CorefValue.Positive:
            if rating is None:
                rating = 1
        else:
            rating = 0

        # Fix ordering of content, subtopic pairs
        if ((content_id2 < content_id1 or
             (content_id2 == content_id1 and subtopic_id2 < subtopic_id1))):
            self.content_id1 = content_id2
            self.subtopic_id1 = subtopic_id2
            self.content_id2 = content_id1
            self.subtopic_id2 = subtopic_id1
        else:
            self.content_id1 = content_id1
            self.subtopic_id1 = subtopic_id1
            self.content_id2 = content_id2
            self.subtopic_id2 = subtopic_id2

        self.annotator_id = annotator_id
        self.value = value
        self.epoch_ticks = epoch_ticks
        self.rating = rating

        self.meta = meta
        if self.meta is None:
            self.meta = {}

    def update(self, **kwargs):
        '''Creates a new label with the modifications given.

        ``kwargs`` supports the same parameters as this class'
        constructor. A new label is created from a copy of
        ``self`` with the parameters given as overrides.

        For example, ``lab.update()`` creates a copy of ``lab``.
        Another example, ``lab.update(annotator_id='foo')``
        creates a copy of ``lab`` except the ``annotator_id`` field
        is set to ``'foo'``.

        (This is like ``namedtuple._replace`` or record updates in
        functional languages.)
        '''
        return Label(**dict(self.as_dict(), **kwargs))

    def as_dict(self):
        '''Returns this label as a Python dictionary.'''
        return {
            'content_id1': self.content_id1, 'subtopic_id1': self.subtopic_id1,
            'content_id2': self.content_id2, 'subtopic_id2': self.subtopic_id2,
            'annotator_id': self.annotator_id, 'value': self.value,
            'epoch_ticks': self.epoch_ticks, 'rating': self.rating,
            'meta': self.meta,
        }

    def __contains__(self, v):
        '''Tests membership of identifiers.

        If `v` is a tuple of ``(content_id, subtopic_id)``, then the
        pair is checked for membership. Otherwise, `v` must be a
        :class:`str` and is checked for equality to one of the content
        ids in this label.

        As a special case, if `v` is ``(content_id, None)``, then
        `v` is treated as if it were ``content_id``.

        >>> l = Label('c1', 'c2', 'a', 1)
        >>> 'c1' in l
        True
        >>> 'a' in l
        False
        >>> ('c1', None) in l
        True
        >>> ('c1', 's1') in l
        False
        >>> ll = Label('c1', 'c2', 'a', 1, 's1', 's2')
        >>> 'c1' in l
        True
        >>> ('c1', None) in l
        True
        >>> ('c1', 's1') in l
        True

        '''
        if isinstance(v, tuple) and len(v) == 2 and v[1] is None:
            v = v[0]

        if isinstance(v, tuple) and len(v) == 2:
            cid, subid = v
            cid1, subid1 = self.content_id1, self.subtopic_id1
            cid2, subid2 = self.content_id2, self.subtopic_id2
            return (cid == cid1 and subid == subid1) \
                or (cid == cid2 and subid == subid2)
        else:
            return v == self.content_id1 or v == self.content_id2

    def other(self, content_id):
        '''Returns the other content id.

        If ``content_id == self.content_id1``, then return
        ``self.content_id2`` (and vice versa).  Raises
        :exc:`exceptions.KeyError` if `content_id` is neither one.

        >>> l = Label('c1, 'c2', 'a', 1)
        >>> l.other('c1')
        'c2'
        >>> l.other('c2')
        'c1'
        >>> l.other('a')
        Traceback (most recent call last):
            ...
        KeyError: 'a'

        '''
        if content_id == self.content_id1:
            return self.content_id2
        elif content_id == self.content_id2:
            return self.content_id1
        else:
            raise KeyError(content_id)

    def subtopic_for(self, content_id):
        '''Get the subtopic id that corresponds with a content id.

        >>> l = Label('c1', 'c2', 'a', 1, 's1', 's2')
        >>> l.subtopic_for('c1')
        's1'
        >>> l.subtopic_for('c2')
        's2'
        >>> l.subtopic_for('a')
        Traceback (most recent call last):
            ...
        KeyError: 'a'

        :param str content_id: content ID to look up
        :return: subtopic ID for `content_id`
        :raise exceptions.KeyError: if `content_id` is neither
          content ID in this label

        '''

        if content_id == self.content_id1:
            return self.subtopic_id1
        elif content_id == self.content_id2:
            return self.subtopic_id2
        else:
            raise KeyError(content_id)

    def __lt__(self, other):
        if self.content_id1 != other.content_id1:
            return self.content_id1 < other.content_id1
        if self.content_id2 != other.content_id2:
            return self.content_id2 < other.content_id2
        if self.subtopic_id1 != other.subtopic_id1:
            return self.subtopic_id1 < other.subtopic_id1
        if self.subtopic_id2 != other.subtopic_id2:
            return self.subtopic_id2 < other.subtopic_id2
        if self.annotator_id != other.annotator_id:
            return self.annotator_id < other.annotator_id
        if self.epoch_ticks != other.epoch_ticks:
            # opposite ordering here (consciously -- newer is smaller)
            return self.epoch_ticks > other.epoch_ticks
        if self.value is not other.value:
            return self.value < other.value
        if self.rating != other.rating:
            return self.rating < other.rating
        return False  # self == other

    def __eq__(self, other):
        if self.content_id1 != other.content_id1:
            return False
        if self.content_id2 != other.content_id2:
            return False
        if self.subtopic_id1 != other.subtopic_id1:
            return False
        if self.subtopic_id2 != other.subtopic_id2:
            return False
        if self.annotator_id != other.annotator_id:
            return False
        if self.epoch_ticks != other.epoch_ticks:
            return False
        if self.value is not other.value:
            return False
        if self.rating != other.rating:
            return False
        return True

    def same_subject_as(self, other):
        '''Determine if two labels are about the same thing.

        This predicate returns :const:`True` if `self` and `other`
        have the same content IDs, subtopic IDs, and annotator ID.
        The other fields may have any value.

        >>> t = time.time()
        >>> l1 = Label('c1', 'c2', 'a', CorefValue.Positive,
        ...            epoch_ticks=t)
        >>> l2 = Label('c1', 'c2', 'a', CorefValue.Negative,
        ...            epoch_ticks=t)
        >>> l1.same_subject_as(l2)
        True
        >>> l1 == l2
        False

        '''
        if self.content_id1 != other.content_id1:
            return False
        if self.content_id2 != other.content_id2:
            return False
        if self.subtopic_id1 != other.subtopic_id1:
            return False
        if self.subtopic_id2 != other.subtopic_id2:
            return False
        if self.annotator_id != other.annotator_id:
            return False
        return True

    def changed(self, other):
        '''Returns true if ``other`` is modified from ``self`.

        More specifically, returns true if and only if ``self``
        and ``other`` share the same subject but have different
        coref/rating values.
        '''
        return self.same_subject_as(other) \
                and (self.value, self.rating) != (other.value, other.rating)

    @staticmethod
    def most_recent(labels):
        '''Filter an iterator to return the most recent for each subject.

        `labels` is any iterator over :class:`Label` objects.  It
        should be sorted with the most recent first, which is the
        natural sort order that :func:`sorted` and the
        :class:`LabelStore` adapter will return.  The result of this
        is a generator of the same labels but with any that are not
        the most recent for the same subject (according to
        :meth:`same_subject_as`) filtered out.

        '''
        prev_label = None
        for label in labels:
            if prev_label is None or not label.same_subject_as(prev_label):
                yield label
            prev_label = label

    @staticmethod
    def diff(old, new):
        '''Returns a *diff* of two sets of labels.

        The diff returned represents the changes required to the set
        of ``old`` labels in order to provide the same semantic meaning
        as the set of ``new`` labels. Thus, the diff is a Python dictionary
        with three keys: ``add``, ``delete`` and ``change``. All three
        keys map to a sequence of labels.

        ``add`` corresponds to the labels that are in ``new`` but not in
        ``old``.

        ``delete`` corresponds to the labels that are in ``old`` but not
        in ``new``.

        ``change`` corresponds to the labels that are in both ``old``
        and ``new`` but have different coref/rating values.
        '''
        # A wrapper type for a more relaxed label equality comparison.
        # (We use a wrapper type so that its easy to unwrap it.)
        class SubjectLabel(object):
            def __init__(self, lab):
                self.lab = lab
            def changed(self, other):
                return self.lab.changed(other.lab)
            def __eq__(self, other):
                return self.lab.same_subject_as(other.lab)
            def __hash__(self):
                return hash((self.lab.content_id1, self.lab.content_id2,
                             self.lab.subtopic_id1, self.lab.subtopic_id2,
                             self.lab.annotator_id))

        old, new = set(map(SubjectLabel, old)), set(map(SubjectLabel, new))
        add, delete = list(new.difference(old)), list(old.difference(new))
        change = map(itemgetter(1),
                     ifilter(lambda (o, n): o.changed(n), product(old, new)))
        unwrap = attrgetter('lab')
        return {
            'add': map(unwrap, add),
            'delete': map(unwrap, delete),
            'change': map(unwrap, change),
        }

    def __hash__(self):
        return (hash(self.content_id1) ^
                hash(self.content_id2) ^
                hash(self.subtopic_id1) ^
                hash(self.subtopic_id2) ^
                hash(self.annotator_id) ^
                hash(self.epoch_ticks) ^
                hash(self.value) ^
                hash(self.rating))

    def __str__(self):
        res = repr(self.content_id1)[1:-1] # strip quotes
        if self.subtopic_id1:
            res += '(' + repr(self.subtopic_id1)[1:-1] + ')'
        if self.value is CorefValue.Positive:
            res += ' =='
        elif self.value is CorefValue.Unknown:
            res += ' ??'
        elif self.value is CorefValue.Negative:
            res += ' !='
        else:
            res += ' **'
        res += '(' + str(self.rating) + ') '
        res += repr(self.content_id2)[1:-1] # strip quotes
        if self.subtopic_id2:
            res += '(' + repr(self.subtopic_id2)[1:-1] + ')'
        res += ' by ' + self.annotator_id
        res += ' at ' + str(datetime.utcfromtimestamp(self.epoch_ticks))
        return res

    def __repr__(self):
        return ('Label('
                'content_id1={0.content_id1}, '
                'content_id2={0.content_id2}, '
                'annotator_id={0.annotator_id}, '
                'value={0.value}, '
                'subtopic_id1={0.subtopic_id1}, '
                'subtopic_id2={0.subtopic_id2}, '
                'epoch_ticks={0.epoch_ticks}, '
                'rating={0.rating})'.format(self))


class LabelStore(object):
    '''A label database.

    .. automethod:: __init__
    .. automethod:: put
    .. automethod:: get
    .. automethod:: directly_connected
    .. automethod:: connected_component
    .. automethod:: expand
    .. automethod:: everything
    .. automethod:: delete_all
    '''
    config_name = 'dossier.label'
    TABLE = 'label'

    _kvlayer_namespace = {
        # (cid1, cid2, subid1, subid2, annotator_id, time) -> value
        # N.B. The `long` type here is for the benefit of the underlying
        # database storage. It will hopefully result in a storage type
        # that is big enough to contain milliseconds epoch ticks. (A 64 bit
        # integer is more than sufficient, which makes `long` unnecessary from
        # a Python 2 perspective.)
        TABLE: (str, str, str, str, str, long),
    }

    def __init__(self, kvlclient):
        '''Create a new label store.

        :param kvlclient: kvlayer client
        :type kvlclient: :class:`kvlayer._abstract_storage.AbstractStorage`
        :rtype: :class:`LabelStore`
        '''
        self.kvl = kvlclient
        self.kvl.setup_namespace(self._kvlayer_namespace)

    def put(self, *labels):
        '''Add a new label to the store.

        :param label: label
        :type label: :class:`Label`
        '''
        puts = []
        for label in labels:
            k1, k2 = self._keys_from_label(label)
            v = self._value_from_label(label)
            puts.append((k1, v))
            puts.append((k2, v))
        self.kvl.put(self.TABLE, *puts)

    def _keys_from_label(self, label):
        # `label` is stored under both normal and swapped key tuples,
        # so that we can efficiently find c2<->c1 labels
        k1 = (label.content_id1, label.content_id2,
              label.subtopic_id1, label.subtopic_id2,
              label.annotator_id, time_complement(label.epoch_ticks))
        k2 = (label.content_id2, label.content_id1,
              label.subtopic_id2, label.subtopic_id1,
              label.annotator_id, time_complement(label.epoch_ticks))
        return k1, k2

    def _value_from_label(self, label):
        # Pack value and rating into a single byte, since both will likely
        # be small integers.
        to_pack = (label.value.value+1) | (label.rating << 4)
        v = struct.pack('B', to_pack)

        # Encode the metadata.
        meta_ser = cbor.dumps(label.meta)

        return v + meta_ser

    def get(self, cid1, cid2, annotator_id, subid1='', subid2=''):
        '''Retrieve a label from the store.

        When ``subid1`` and ``subid2`` are empty, then a label without
        subtopic identifiers will be returned.

        If there are multiple labels stored with the same parts,
        the single most recent one will be returned.  If there are
        no labels with these parts, :exc:`exceptions.KeyError` will
        be raised.

        :param str cid1: content id
        :param str cid2: content id
        :param str annotator_id: annotator id
        :param str subid1: subtopic id
        :param str subid2: subtopic id
        :rtype: :class:`Label`
        :raises: :exc:`KeyError` if no label could be found.

        '''
        t = (cid1, cid2, subid1, subid2, annotator_id)

        # We return the first result because the `kvlayer` abstraction
        # guarantees that the first result will be the most recent entry
        # for this particular key (since the timestamp is inserted as a
        # complement value).
        for k, v in self.kvl.scan(self.TABLE, (t, t)):
            return self._label_from_kvlayer(k, v)
        raise KeyError(t)

    def _label_from_kvlayer(self, k, v):
        '''Make a label from a kvlayer row.'''
        (content_id1, content_id2, subtopic_id1, subtopic_id2,
         annotator_id, inverted_epoch_ticks) = k
        epoch_ticks = time_complement(inverted_epoch_ticks)

        (unpacked,) = struct.unpack('B', v[0])
        value = (unpacked & 15) - 1
        rating = (unpacked >> 4)
        meta = None
        if len(v) > 1:
            meta = cbor.loads(v[1:])

        return Label(content_id1=content_id1,
                     content_id2=content_id2,
                     annotator_id=annotator_id,
                     value=value,
                     subtopic_id1=subtopic_id1,
                     subtopic_id2=subtopic_id2,
                     epoch_ticks=epoch_ticks,
                     rating=rating,
                     meta=meta)

    def directly_connected(self, ident):
        '''Return a generator of labels connected to ``ident``.

        ``ident`` may be a ``content_id`` or a ``(content_id,
        subtopic_id)``.

        If no labels are defined for ``ident``, then the generator
        will yield no labels.

        Note that this only returns *directly* connected labels. It
        will not follow transitive relationships.

        :param ident: content id or (content id and subtopic id)
        :type ident: ``str`` or ``(str, str)``
        :rtype: generator of :class:`Label`
        '''
        content_id, subtopic_id = normalize_ident(ident)
        return self.everything(include_deleted=False,
                               content_id=content_id,
                               subtopic_id=subtopic_id)

    def connected_component(self, ident):
        '''Return a connected component generator for ``ident``.

        ``ident`` may be a ``content_id`` or a ``(content_id,
        subtopic_id)``.

        Given an ``ident``, return the corresponding connected
        component by following all positive transitivity relationships.

        For example, if ``(a, b, 1)`` is a label and ``(b, c, 1)`` is
        a label, then ``connected_component('a')`` will return both
        labels even though ``a`` and ``c`` are not directly connected.

        (Note that even though this returns a generator, it will still
        consume memory proportional to the number of labels in the
        connected component.)

        :param ident: content id or (content id and subtopic id)
        :type ident: ``str`` or ``(str, str)``
        :rtype: generator of :class:`Label`
        '''
        ident = normalize_ident(ident)
        done = set()  # set of cids that we've queried with
        todo = set([ident])  # set of cids to do a query for
        labels = set()
        while todo:
            ident = todo.pop()
            done.add(ident)
            for label in self.directly_connected(ident):
                if label.value != CorefValue.Positive:
                    continue
                ident1, ident2 = idents_from_label(
                    label, subtopic=ident_has_subtopic(ident))
                if ident1 not in done:
                    todo.add(ident1)
                if ident2 not in done:
                    todo.add(ident2)

                if label not in labels:
                    labels.add(label)
                    yield label

    def expand(self, ident):
        '''Return expanded set of labels from a connected component.

        The connected component is derived from ``ident``. ``ident``
        may be a ``content_id`` or a ``(content_id, subtopic_id)``.
        If ``ident`` identifies a subtopic, then expansion is done
        on a subtopic connected component (and expanded labels retain
        subtopic information).

        The labels returned by :meth:`LabelStore.connected_component`
        contains only the :class:`Label` stored in the
        :class:`LabelStore`, and does not include the labels you can
        infer from the connected component. This method returns both
        the data-backed labels and the inferred labels.

        Subtopic assignments of the expanded labels will be empty. The
        ``annotator_id`` will be an arbitrary ``annotator_id`` within
        the connected component.

        :param str content_id: content id
        :param value: coreferent value
        :type value: :class:`CorefValue`
        :rtype: ``list`` of :class:`Label`
        '''
        subtopic = ident_has_subtopic(normalize_ident(ident))
        labels = list(self.connected_component(ident))
        labels.extend(expand_labels(labels, subtopic=subtopic))
        return labels

    def negative_inference(self, content_id):
        '''Return a generator of inferred negative label relationships
        centered on ``content_id``.

        Negative labels are inferred by getting all other content ids
        connected to ``content_id`` through a negative label, then
        running :meth:`LabelStore.negative_label_inference` on those
        labels. See :meth:`LabelStore.negative_label_inference` for
        more information.
        '''
        neg_labels = ifilter(lambda l: l.value == CorefValue.Negative,
                             self.directly_connected(content_id))
        for label in neg_labels:
            label_inf = self.negative_label_inference(label)
            for label in label_inf:
                yield label

    def negative_label_inference(self, label):
        '''Return a generator of inferred negative label relationships.

        Construct ad-hoc negative labels between ``label.content_id1``
        and the positive connected component of ``label.content_id2``,
        and ``label.content_id2`` to the connected component of
        ``label.content_id1``.

        Note this will allocate memory proportional to the size of the
        connected components of ``label.content_id1`` and
        ``label.content_id2``.
        '''
        assert label.value == CorefValue.Negative

        yield label

        cid2_comp = self.connected_component(label.content_id2)
        for comp_label in cid2_comp:
            comp_cids = (comp_label.content_id1,
                         comp_label.content_id2)
            for comp_cid in comp_cids:
                if not comp_cid == label.content_id2:
                    yield Label(label.content_id1, comp_cid,
                                'auto', CorefValue.Negative)

        cid1_comp = self.connected_component(label.content_id1)
        for comp_label in cid1_comp:
            comp_cids = (comp_label.content_id1,
                         comp_label.content_id2)
            for comp_cid in comp_cids:
                if not comp_cid == label.content_id1:
                    yield Label(label.content_id2, comp_cid,
                                'auto', CorefValue.Negative)

    def _filter_keys(self, content_id=None, subtopic_id=None):
        '''Filter out-of-order labels by key tuple.

        :class:`Label` always sorts by `(cid1,cid2,sid1,sid2)`, but
        for efficient lookups on `cid2` this class also stores in
        order `(cid2,cid1,sid2,sid1)`.  Filter out things that are
        in the wrong order.  But, if an original query specified
        `content_id` or `subtopic_id`, account for the possibility
        that something might be apparently out-of-order but its
        dual will not be in the query at all.

        '''
        def accept(kvp):
            (content_id1, content_id2, subtopic_id1, subtopic_id2,
             annotator_id, inverted_epoch_ticks) = kvp[0]
            if content_id is None:
                # We're scanning everything, so accept the label if
                # it's the natural order; l.content_id1 == cid1
                return (content_id1 < content_id2 or
                        (content_id1 == content_id2 and
                         subtopic_id1 <= subtopic_id2))
            assert content_id1 == content_id  # because that's the scan range
            # If we're not looking for subtopic IDs, then accept records
            # matching the content ID that are in natural order
            if subtopic_id is None:
                if content_id2 != content_id:
                    return True  # will never see its dual
                return subtopic_id1 <= subtopic_id2
            # The scan range doesn't include subtopic IDs (the key schema
            # is oriented towards querying content-to-content labels)
            # so we need to accept records where either part matches
            # (if both parts match then the record is its own dual)
            if subtopic_id == subtopic_id1:
                return True
            if content_id == content_id2 and subtopic_id == subtopic_id2:
                return True
            return False
        return accept

    def everything(self, include_deleted=False, content_id=None,
                   subtopic_id=None):
        '''Returns a generator of all labels in the store.

        If `include_deleted` is :const:`True`, labels that have been
        overwritten with more recent labels are also included.  If
        `content_id` is not :const:`None`, only labels for that
        content ID are retrieved; and then if `subtopic_id` is not
        :const:`None`, only that subtopic is retrieved, else all
        subtopics are retrieved.  The returned labels will always be
        in sorted order, content IDs first, and with those with the
        same content, subtopic, and annotator IDs sorted newest first.

        :rtype: generator of :class:`Label`

        '''
        if content_id is not None:
            ranges = [((content_id,), (content_id,))]
        else:
            ranges = []
        labels = self.kvl.scan(self.TABLE, *ranges)
        labels = ifilter(self._filter_keys(content_id, subtopic_id), labels)
        labels = imap(lambda p: self._label_from_kvlayer(*p), labels)
        if not include_deleted:
            labels = Label.most_recent(labels)
        return labels

    def delete_all(self):
        '''Deletes all labels in the store.'''
        self.kvl.clear_table(self.TABLE)

    def delete(self, *labels):
        '''Delete `labels` from the store, which involves deleting two
        records for each :class:`Label`.

        :rtype: None
        :raises: :exc:`KeyError` if any of the `labels` could not be found.

        '''
        deletes = []
        for label in labels:
            k1, k2 = self._keys_from_label(label)
            deletes.append(k1)
            deletes.append(k2)
        self.kvl.delete(self.TABLE, *deletes)

    def apply_diff(self, diff):
        '''Applies a diff to the label table.

        A ``diff`` is a dictionary with three keys: ``add``, ``delete``
        and ``change``. Each key should map to a list of labels.

        ``add`` corresponds to the labels that are in ``new`` but not in
        ``old``.

        ``delete`` corresponds to the labels that are in ``old`` but not
        in ``new``.

        ``change`` corresponds to the labels that are in both ``old``
        and ``new`` but have different coref/rating values.
        '''
        # add and change are easy---just put the labels.
        # delete is a little trickier. We need to scrub out the impact of
        # the previous label without actually deleting it. For this, we
        # use an unknown coref value.
        insert = (
            diff['add']
            + diff['change']
            + [lab.update(value=CorefValue.Unknown) for lab in diff['delete']]
        )
        self.put(*insert)


def unordered_pair_eq(pair1, pair2):
    '''Performs pairwise unordered equality.

    ``pair1`` == ``pair2`` if and only if
    ``frozenset(pair1)`` == ``frozenset(pair2)``.
    '''
    (x1, y1), (x2, y2) = pair1, pair2
    return (x1 == x2 and y1 == y2) or (x1 == y2 and y1 == x2)


def normalize_pair(x, y):
    '''Normalize a pair of values.

    Returns ``(y, x)`` if and only if ``y < x`` and ``(x, y)``
    otherwise.
    '''
    if y < x:
        return y, x
    else:
        return x, y


def expand_labels(labels, subtopic=False):
    '''Expand a set of labels that define a connected component.

    ``labels`` must define a *positive* connected component: it is all
    of the edges that make up the *single* connected component in the
    :class:`LabelStore`. expand will ignore subtopic assignments, and
    annotator_id will be an arbitrary one selected from ``labels``.

    Note that this function only returns the expanded labels, which
    is guaranteed to be disjoint with the given ``labels``. This
    requirement implies that ``labels`` is held in memory to ensure
    that no duplicates are returned.

    If ``subtopic`` is ``True``, then it is assumed that ``labels``
    defines a ``subtopic`` connected component. In this case, subtopics
    are included in the expanded labels.

    :param labels: iterable of :class:`Label` for the connected component.
    :rtype: generator of expanded :class:`Label`s only
    '''
    labels = list(labels)
    assert all(lab.value == CorefValue.Positive for lab in labels)

    # Anything to expand?
    if len(labels) == 0:
        return

    annotator = labels[0].annotator_id

    data_backed = set()
    connected_component = set()
    for label in labels:
        ident1, ident2 = idents_from_label(label, subtopic=subtopic)
        data_backed.add(normalize_pair(ident1, ident2))
        connected_component.add(ident1)
        connected_component.add(ident2)

    # We do not want to rebuild the Labels we already have,
    # because they have true annotator_id and subtopic
    # fields that we may want to preserve.
    for ident1, ident2 in combinations(connected_component, 2):
        if normalize_pair(ident1, ident2) not in data_backed:
            (cid1, subid1), (cid2, subid2) = ident1, ident2
            yield Label(cid1, cid2, annotator, CorefValue.Positive,
                        subtopic_id1=subid1, subtopic_id2=subid2)


def normalize_ident(ident):
    '''Splits a generic identifier.

    If ``ident`` is a tuple, then ``(ident[0], ident[1])`` is returned.
    Otherwise, ``(ident[0], None)`` is returned.
    '''
    if isinstance(ident, tuple) and len(ident) == 2:
        return ident[0], ident[1]  # content_id, subtopic_id
    else:
        return ident, None


def ident_has_subtopic(ident):
    return ident[1] is not None


def idents_from_label(lab, subtopic=False):
    '''Returns the "ident" of a label.

    If ``subtopic`` is ``True``, then a pair of pairs is returned,
    where each pair corresponds to the content id and subtopic id in
    the given label.

    Otherwise, a pair of pairs is returned, but the second element of each
    pair is always ``None``.

    This is a helper function that is useful for dealing with generic
    label identifiers.
    '''
    if not subtopic:
        return (lab.content_id1, None), (lab.content_id2, None)
    else:
        return (
            (lab.content_id1, lab.subtopic_id1),
            (lab.content_id2, lab.subtopic_id2),
        )



'''A simple storage interface for labels (ground truth data).

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2014 Diffeo, Inc.

:mod:`dossier.label` provides a convenient interface to a
:mod:`kvlayer` table for storing ground truth data, otherwise known as
"labels." Each label, at the highest level, maps two things (addressed
by content identifiers) to a coreferent value. This coreferent value is
an indication by a human that these two things are "the same", "not the
same" or "I don't know if they are the same." *Sameness* in this case
is determined by the human doing the annotation.

Each label also contains an ``annotator_id``, which identifies the
human that created the label. A timestamp (in milliseconds since the
Unix epoch) is also included on every label.

Example
-------
Using a storage backend in your code requires a working ``kvlayer``
configuration, which is usually written in a YAML file like so:

.. code-block:: yaml

    kvlayer:
      app_name: store
      namespace: dossier
      storage_type: redis
      storage_addresses: ["redis.example.com:6379"]

And here's a full working example that uses local memory to store
labels:

.. code-block:: python

    from dossier.label import Label, LabelStore, CorefValue
    import kvlayer
    import yakonfig

    yaml = """
    kvlayer:
      app_name: store
      namespace: dossier
      storage_type: local
    """
    with yakonfig.defaulted_config([kvlayer], yaml=yaml):
        label_store = LabelStore(kvlayer.client())

        lab = Label('a', 'b', 'annotator', CorefValue.Positive)
        label_store.put(lab)

        assert lab == label_store.get('a', 'b', 'annotator')

See the documentation for :mod:`yakonfig` for more details on the
configuration setup.

.. autoclass:: LabelStore
.. autoclass:: Label
.. autoclass:: CorefValue

:command:`dossier.label` command-line tool
==========================================

.. automodule:: dossier.label.run
'''
from __future__ import absolute_import, division, print_function

from dossier.label.label import Label, LabelStore, CorefValue, expand_labels

__all__ = ['Label', 'LabelStore', 'CorefValue', 'expand_labels']



'''dossier.label.tests

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

from pyquchk.arbitraries import int_, str_letters
import pytest

import kvlayer
import yakonfig


# Generators of random values.
coref_value = int_(low=-1, high=1, uniform=True)
# "Sometime in 2014"
time_value = int_(low=1388534400, high=1420070400, uniform=True)
id_ = str_letters


@pytest.yield_fixture
def kvl():
    config = {
        'storage_type': 'local',
        'app_name': 'diffeo',
        'namespace': 'dossier.label.tests',
    }
    with yakonfig.defaulted_config([kvlayer], params=config) as config:
        client = kvlayer.client()
        yield client
        client.delete_namespace()
        client.close()



'''dossier.label.tests

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.
'''
from __future__ import absolute_import, division, print_function

from pyquchk import qc
import pytest
import struct

from dossier.label import Label, LabelStore
from dossier.label.tests import kvl, coref_value, time_value, id_  # noqa


@pytest.yield_fixture  # noqa
def label_store(kvl):
    lstore = LabelStore(kvl)
    yield lstore
    lstore.delete_all()


def test_no_prefix(label_store):
    foo_bar = Label('Foo', 'Bar', '', 1)
    foobaz_bar = Label('Foo Baz', 'Bar', '', 1)

    label_store.put(foo_bar)
    label_store.put(foobaz_bar)

    direct = list(label_store.directly_connected('Foo'))
    assert direct == [foo_bar]


def test_no_prefix_subtopic(label_store):
    foo_bar = Label('Foo', 'Bar', '', 1, 'Foo', 'Bar')
    foobaz_bar = Label('Foo Baz', 'Bar', '', 1, 'Foo Baz', 'Bar')

    label_store.put(foo_bar)
    label_store.put(foobaz_bar)

    direct = list(label_store.directly_connected(('Foo', 'Foo')))
    assert direct == [foo_bar]


def test_put_get(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value):
        label_store.delete_all()

        lab = Label(cid1, cid2, ann, v)
        label_store.put(lab)
        got = label_store.get(cid1, cid2, ann)
        assert lab == got and lab.value == got.value
    _()


def test_put_get_delete_get(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value):
        label_store.delete_all()

        lab = Label(cid1, cid2, ann, v)
        label_store.put(lab)
        got = label_store.get(cid1, cid2, ann)
        assert lab == got and lab.value == got.value

        label_store.delete(lab)
        with pytest.raises(KeyError):
            label_store.get(cid1, cid2, ann)
    _()


def test_put_get_unordered(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value):
        label_store.delete_all()

        lab = Label(cid1, cid2, ann, v)
        label_store.put(lab)
        got = label_store.get(cid2, cid1, ann)
        assert lab == got and lab.value == got.value
    _()


def test_put_get_recent(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)
        got = label_store.get(cid1, cid2, ann)
        assert got == lab2
        assert got != lab1
        assert got.value == lab2.value
    _()


def test_put_get_recent_unordered(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid2, cid1, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)
        got = label_store.get(cid1, cid2, ann)
        assert got == lab2
        assert got != lab1
        assert got.value == lab2.value
    _()


def test_everything_simple(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value):
        label_store.delete_all()
        l = Label(cid1, cid2, ann, v)
        label_store.put(l)
        assert list(label_store.everything()) == [l]
    _()


def test_everything_two(label_store):
    @qc
    def _(cid1a=id_, cid1b=id_, ann1=id_, v1=coref_value, t1=time_value,
          cid2a=id_, cid2b=id_, ann2=id_, v2=coref_value, t2=time_value):
        label_store.delete_all()
        l1 = Label(cid1a, cid1b, ann1, v1, epoch_ticks=t1)
        l2 = Label(cid2a, cid2b, ann2, v2, epoch_ticks=t2)
        label_store.put(l1)
        label_store.put(l2)
        if l1.same_subject_as(l2) and l1.epoch_ticks == l2.epoch_ticks:
            expected = [l2]
        else:
            expected = list(sorted([l1, l2]))
        assert (list(label_store.everything(include_deleted=True)) ==
                expected)
    _()


def test_everything_overwrite(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v=coref_value, t=time_value):
        label_store.delete_all()
        l1 = Label(cid1, cid2, ann, v, epoch_ticks=t)
        l2 = Label(cid1, cid2, ann, v, epoch_ticks=t+1)
        label_store.put(l1)
        label_store.put(l2)
        assert list(label_store.everything()) == [l2]
        assert list(label_store.everything(include_deleted=True)) == [l2, l1]
    _()


def test_everything_content_id(label_store):
    @qc
    def _(cid1a=id_, cid1b=id_, ann1=id_, v1=coref_value, t1=time_value,
          cid2a=id_, cid2b=id_, ann2=id_, v2=coref_value, t2=time_value):
        label_store.delete_all()
        l1 = Label(cid1a, cid1b, ann1, v1, epoch_ticks=t1)
        l2 = Label(cid2a, cid2b, ann2, v2, epoch_ticks=t2)
        label_store.put(l1)
        label_store.put(l2)
        if l1.same_subject_as(l2):
            if l1.epoch_ticks == l2.epoch_ticks:
                expected = [l2]
            else:
                expected = list(sorted([l1, l2]))[0:1]
        elif cid1a == cid2a or cid1a == cid2b:
            expected = list(sorted([l1, l2]))
        else:
            expected = [l1]
        assert (list(label_store.everything(content_id=cid1a)) == expected)
    _()


def test_everything_subtopic_id(label_store):
    @qc
    def _(cid1a=id_, cid1b=id_, sid1a=id_, sid1b=id_, ann1=id_,
          v1=coref_value, t1=time_value,
          cid2a=id_, cid2b=id_, sid2a=id_, sid2b=id_, ann2=id_,
          v2=coref_value, t2=time_value):
        label_store.delete_all()
        l1 = Label(cid1a, cid1b, ann1, v1, epoch_ticks=t1,
                   subtopic_id1=sid1a, subtopic_id2=sid1b)
        l2 = Label(cid2a, cid2b, ann2, v2, epoch_ticks=t2,
                   subtopic_id1=sid2a, subtopic_id2=sid2b)
        label_store.put(l1)
        label_store.put(l2)
        if l1.same_subject_as(l2):
            if l1.epoch_ticks == l2.epoch_ticks:
                expected = [l2]
            else:
                expected = list(sorted([l1, l2]))[0:1]
        elif ((cid1a == cid2a and sid1a == sid2a) or
              (cid1a == cid2b and sid1a == sid2b)):
            expected = list(sorted([l1, l2]))
        else:
            expected = [l1]
        assert (list(label_store.everything(content_id=cid1a,
                                            subtopic_id=sid1a)) == expected)
    _()


def test_direct_connect_recent(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid1, cid2, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)

        assert list(label_store.directly_connected(cid1)) == [lab2]
        assert list(label_store.directly_connected(cid2)) == [lab2]
    _()


def test_direct_connect_recent_unordered(label_store):
    @qc
    def _(cid1=id_, cid2=id_, ann=id_, v1=coref_value, v2=coref_value):
        label_store.delete_all()

        lab1 = Label(cid1, cid2, ann, v1)
        lab2 = Label(cid2, cid1, ann, v2, epoch_ticks=lab1.epoch_ticks + 1)
        label_store.put(lab1)
        label_store.put(lab2)

        assert list(label_store.directly_connected(cid1)) == [lab2]
        assert list(label_store.directly_connected(cid2)) == [lab2]
    _()


def test_direct_connect(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    direct = list(label_store.directly_connected('a'))
    assert direct == [ab, ac]


def test_direct_connect_unordered(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('c', 'a', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    direct = list(label_store.directly_connected('a'))
    assert direct == [ab, ac]


def test_connected_component_basic(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, ac, bc])


def test_connected_component_unordered(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('c', 'a', '', 1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, ac, bc])


def test_connected_component_diff_value(label_store):
    ab = Label('a', 'b', '', 1)
    ac = Label('a', 'c', '', -1)
    bc = Label('b', 'c', '', 1)
    label_store.put(ab)
    label_store.put(ac)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc])


def test_connected_component_many(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', 1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc, cd])


def test_connected_component_many_diff_value(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', -1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab])


def test_connected_component_many_most_recent(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', -1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab])

    # This label should overwrite the existing `bc` label and expand
    # the connected component to `cd` through transitivity.
    bc = Label('b', 'c', '', 1, epoch_ticks=bc.epoch_ticks + 1)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc, cd])


def test_connected_component_many_most_recent_diff_value(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', 1)
    cd = Label('c', 'd', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab, bc, cd])

    # This label should overwrite the existing `bc` label and contract
    # the connected component to just `ab`.
    bc = Label('b', 'c', '', -1, epoch_ticks=bc.epoch_ticks + 1)
    label_store.put(bc)

    connected = list(label_store.connected_component('a'))
    assert frozenset(connected) == frozenset([ab])


def test_connected_component_collision(label_store):
    # You can't store the hashes of objects and expect there to never
    # be collisions.  As a corollary, hash(str) isn't that great
    # vs. small changes, and the recommended technique of xoring
    # together field hashes can get collisions quickly.
    # In particular, hash('test0') ^ hash('test1') is 1,
    # as is hash('test2') ^ hash('test3').
    ab = Label('test0', 'test1', '', 1)
    bc = Label('test1', 'test2', '', 1)
    cd = Label('test2', 'test3', '', 1)
    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)

    assert list(label_store.connected_component('test0')) == [ab, bc, cd]


def test_expand(label_store):
    ab = Label('a', 'b', '', 1)
    bc = Label('b', 'c', '', 1)
    cd = Label('c', 'd', '', 1)
    ae = Label('a', 'e', '', -1)
    fg = Label('f', 'g', '', 1)

    label_store.put(ab)
    label_store.put(bc)
    label_store.put(cd)
    label_store.put(ae)
    label_store.put(fg)

    correct_pairs = [Label('a', 'b', '', 1),
                     Label('a', 'c', '', 1),
                     Label('a', 'd', '', 1),
                     Label('b', 'c', '', 1),
                     Label('b', 'd', '', 1),
                     Label('c', 'd', '', 1)]

    assert frozenset(label_store.expand('a')) == frozenset(correct_pairs)
    assert len(label_store.expand('e')) == 0
    assert label_store.expand('f') == [Label('f', 'g', '', 1)]


def test_negative_label_inference(label_store):
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)

    de = Label('d', 'e', '', 1)
    df = Label('d', 'f', '', 1)
    dg = Label('d', 'g', '', -1)

    cd = Label('c', 'd', '', -1)
    fh = Label('f', 'h', '', 1)

    label_store.put(ac)
    label_store.put(bc)
    label_store.put(de)
    label_store.put(df)
    label_store.put(cd)
    label_store.put(dg)
    label_store.put(fh)

    def get_pair(label):
        return (label.content_id1, label.content_id2)

    correct_pairs = [('a', 'd'),
                     ('b', 'd'),
                     ('c', 'd'),
                     ('c', 'e'),
                     ('c', 'f'),
                     ('c', 'h')]
    # [but not (a,b) <-/-> (e,f)]

    inference = label_store.negative_label_inference(cd)

    assert frozenset(map(get_pair, inference)) == \
        frozenset(correct_pairs)


def test_negative_inference(label_store):
    ac = Label('a', 'c', '', 1)
    bc = Label('b', 'c', '', 1)

    de = Label('d', 'e', '', 1)
    df = Label('d', 'f', '', 1)

    cg = Label('c', 'g', '', -1)
    dg = Label('d', 'g', '', -1)

    hg = Label('h', 'g', '', 1)

    label_store.put(ac)
    label_store.put(bc)
    label_store.put(de)
    label_store.put(df)
    label_store.put(cg)
    label_store.put(dg)
    label_store.put(hg)

    def get_pair(label):
        return (label.content_id1, label.content_id2)

    correct_pairs = [('a', 'g'),
                     ('b', 'g'),
                     ('c', 'g'),
                     ('c', 'h'),
                     ('d', 'g'),
                     ('d', 'h'),
                     ('e', 'g'),
                     ('f', 'g')]

    inference = label_store.negative_inference('g')

    assert frozenset(map(get_pair, inference)) == \
        frozenset(correct_pairs)


# Subtopic testing is below.


def test_sub_direct_connect(label_store):
    a1b2 = Label('a', 'b', '', 1, '1', '2')
    a1c3 = Label('a', 'c', '', 1, '1', '3')
    b2c3 = Label('b', 'c', '', 1, '2', '3')
    a4b2 = Label('a', 'b', '', 1, '4', '2')
    label_store.put(a1b2)
    label_store.put(a1c3)
    label_store.put(b2c3)
    label_store.put(a4b2)

    # a4b2 should not be included because we're demanding a specific
    # subtopic_id of 'a'.
    direct = list(label_store.directly_connected(('a', '1')))
    assert direct == [a1b2, a1c3]


def test_sub_connected(label_store):
    a1b2 = Label('a', 'b', '', 1, '1', '2')
    b2c3 = Label('b', 'c', '', 1, '2', '3')
    b4c5 = Label('b', 'c', '', 1, '4', '5')
    label_store.put(a1b2)
    label_store.put(b2c3)
    label_store.put(b4c5)

    connected = list(label_store.connected_component(('a', '1')))
    assert frozenset(connected) == frozenset([a1b2, b2c3])


def test_sub_expand(label_store):
    a1b2 = Label('a', 'b', '', 1, '1', '2')
    b2c3 = Label('b', 'c', '', 1, '2', '3')
    b4c5 = Label('b', 'c', '', 1, '4', '5')  # not in subtopic expansion!
    label_store.put(a1b2)
    label_store.put(b2c3)
    label_store.put(b4c5)

    # Not phyiscally present in the label table, but part of expansion!
    a1c3 = Label('a', 'c', '', 1, '1', '3')

    connected = list(label_store.expand(('a', '1')))
    assert frozenset(connected) == frozenset([a1b2, b2c3, a1c3])


# Metadata testing below.


def test_store_legacy_compatibility(label_store):
    def legacy_put_label(label):
        k1, k2 = label_store._keys_from_label(label)
        to_pack = (label.value.value+1) | (label.rating << 4)
        v = struct.pack('B', to_pack)
        label_store.kvl.put(label_store.TABLE, *[(k1, v), (k2, v)])

    label = Label('a', 'b', '', 1, '1', '2')
    legacy_put_label(label)
    label_from_store = label_store.get('a', 'b', '', subid1='1', subid2='2')
    assert label == label_from_store


def test_meta_storage(label_store):
    label = Label('a', 'b', '', 1, '1', '2')
    label.meta['hello'] = 'world'
    label.meta['subtopic1_name'] = 'foo'
    label.meta['some_num'] = 5
    label.meta['some_datastructure'] = [1, 2, 3]

    label_store.put(label)
    label_from_store = label_store.get('a', 'b', '', subid1='1', subid2='2')
    assert label == label_from_store
    assert label.meta == label_from_store.meta



''':command:`dossier.label` command-line tool.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

``dossier.label`` is a command line application for viewing the raw label
data inside the database. Generally, this is a debugging tool for developers.

Run ``dossier.label --help`` for the available commands.
'''
from __future__ import absolute_import, division, print_function
import argparse
from itertools import groupby, imap, islice

import cbor
import kvlayer
import yakonfig

from dossier.label import CorefValue, Label, LabelStore


def label_to_dict(lab):
    return {
        u'content_id1': lab.content_id1,
        u'content_id2': lab.content_id2,
        u'annotator_id': lab.annotator_id,
        u'value': lab.value.value,
        u'subtopic_id1': lab.subtopic_id1,
        u'subtopic_id2': lab.subtopic_id2,
        u'epoch_ticks': lab.epoch_ticks,
        u'rating': lab.rating,
    }


def dict_to_label(d):
    return Label(
        content_id1=d['content_id1'],
        content_id2=d['content_id2'],
        annotator_id=d['annotator_id'],
        value=CorefValue(d['value']),
        subtopic_id1=d.get('subtopic_id1', None),
        subtopic_id2=d.get('subtopic_id2', None),
        epoch_ticks=d.get('epoch_ticks', None),  # will become time.time()
        rating=d.get('rating', None),
    )


def batch(it, size):
    b = []
    for i, v in enumerate(it, 1):
        if i % size == 0:
            yield b
            b = []
        b.append(v)
    if len(b) > 0:
        yield b


class App(yakonfig.cmd.ArgParseCmd):
    def __init__(self, *args, **kwargs):
        yakonfig.cmd.ArgParseCmd.__init__(self, *args, **kwargs)
        self._label_store = None

    @property
    def label_store(self):
        if self._label_store is None:
            self._label_store = LabelStore(kvlayer.client())
        return self._label_store

    def args_list(self, p):
        p.add_argument('--include-deleted', action='store_true',
                       help='When set, show deleted labels.')
        p.add_argument('--substring', help='Case-sensitive string to filter content_ids')
        p.add_argument('--ignore-case', action='store_true',
                       help='Switch substring filtering to case-insensitive')
        p.add_argument('--delete', action='store_true',
                       help='Delete labels that are listed; you will be asked'
                       ' to approve the displayed list.')

    def do_list(self, args):
        labels = self.label_store.everything(
            include_deleted=args.include_deleted)
        def substring_match(k):
            if (args.substring in k.content_id1 or
                args.substring in k.content_id2):
                return True
            if args.ignore_case:
                if (args.substring.lower() in k.content_id1.lower() or
                    args.substring.lower() in k.content_id2.lower()):
                    return True
        to_delete = []
        for k, group in groupby(sorted(labels)):
            if args.substring and not substring_match(k):
                continue
            self.stdout.write('%s\n' % (k,))
            if args.delete:
                to_delete.append(k)
            for lab in islice(group, 1, None):
                self.stdout.write('    %s\n' % repr(lab))
                if args.delete:
                    to_delete.append(lab)
        if args.delete:
            response = raw_input('Are you sure that you want to delete %d labels? [y/N]: ' %
                                 len(to_delete))
            if response.lower() in ['y', 'yes']:
                self.label_store.delete(*to_delete)
                print('Deleted all %d labels.' % len(to_delete))


    def args_dump_all(self, p):
        p.add_argument('--exclude-deleted', action='store_true',
                       help='When set, only the most recent labels are '
                            'dumped.')

    def do_dump_all(self, args):
        labels = list(imap(label_to_dict, self.label_store.everything(
            include_deleted=not args.exclude_deleted)))
        print(cbor.dumps(labels), file=self.stdout)

    def args_show_all(self, p):
        p.add_argument('--exclude-deleted', action='store_true',
                       help='When set, only the most recent labels are '
                            'dumped.')

    def do_show_all(self, args):
        labels = list(imap(label_to_dict, self.label_store.everything(
            include_deleted=not args.exclude_deleted)))
        for label in labels:
            print('%r' % label)

    def args_load(self, p):
        p.add_argument('fpath', nargs='?', default=None,
                       help='File path containing label data. When absent, '
                            'stdin is used.')

    def do_load(self, args):
        if args.fpath is None:
            self._load(self.stdin)
        else:
            with open(args.fpath, 'r') as fp:
                self._load(fp)

    def _load(self, fp):
        for batch_records in batch(cbor.loads(fp.read()), 500):
            labels = map(dict_to_label, batch_records)
            self.label_store.put(*labels)

    def args_get(self, p):
        p.add_argument('content_id', type=str,
                       help='Show all labels directly ascribed to content_id')
        p.add_argument('--value', type=int, choices=[-1, 0, 1],
                       default=None,
                       help='Only show labels with this coreferent value.')

    def do_get(self, args):
        '''Get labels directly connected to a content item.'''
        for label in self.label_store.directly_connected(args.content_id):
            if args.value is None or label.value.value == args.value:
                self.stdout.write('{0}\n'.format(label))

    def args_connected(self, p):
        p.add_argument('content_id', type=str,
                       help='Show all labels connected to content_id')

    def do_connected(self, args):
        '''Find a connected component from positive labels on an item.'''
        connected = self.label_store.connected_component(args.content_id)
        for label in connected:
            self.stdout.write('{0}\n'.format(label))

    def args_delete_all(self, p):
        pass

    def do_delete_all(self, args):
        answer = raw_input('Are you absolutely sure? [y/n] ').strip().lower()
        if answer and answer[0] == 'y':
            self.label_store.delete_all()


def main():
    p = argparse.ArgumentParser(
        description='Interact with DossierStack truth data.')
    app = App()
    app.add_arguments(p)
    args = yakonfig.parse_args(p, [kvlayer, yakonfig])
    app.main(args)



'''Some benchmarks for label construction.

This code demonstrates the performance differences between the
different choices of label representation. Notably, the current
representation is the slowest (by an order of magnitude compared
against the fastest, which is just a raw tuple).

The second fastest is a plain namedtuple with nothing going on
in its constructor. Apparently, we're paying dearly for all the
case analysis in ``dossier.label.Label.__new__``.

This performance comparison may be important some day because label
filtering is done *after* the construction of label objects, which
means we're being wasteful. (But it's oh-so convenient!)

To run the benchmarks, use ``python bench_label_objects.py``.
'''
from __future__ import absolute_import, division, print_function

from collections import namedtuple
import time
from timeit import timeit

from dossier.label import Label as RealLabel, CorefValue

cid, subid, ann_id, ticks = 'cid', 'subid', 'annid', int(time.time() * 1000)
coref_value = CorefValue.Positive


Label = namedtuple('_Label', 'content_id1 content_id2 subtopic_id1 ' +
                             'subtopic_id2 annotator_id epoch_ticks '+
                             'value')

class LabelSlots(object):
    __slots__ = ['cid1', 'cid2', 'subid1', 'subid2', 'ann_id', 'ticks', 'val']
    def __init__(self, cid1, cid2, ann_id, v, subid1, subid2, ticks):
        self.cid1, self.cid2 = cid1, cid2
        self.subid1, self.subid2 = subid1, subid2
        self.ann_id, self.val, self.ticks = ann_id, v, ticks


def bench_slots():
    for _ in xrange(100):
        LabelSlots(cid, cid, ann_id, coref_value, subid, subid, ticks)


def bench_real_namedtuple():
    for _ in xrange(100):
        RealLabel(cid, cid, ann_id, coref_value, subid, subid, ticks)


def bench_namedtuple():
    for _ in xrange(100):
        Label(cid, cid, ann_id, coref_value, subid, subid, ticks)


def bench_tuple():
    for _ in xrange(100):
        (cid, cid, ann_id, coref_value, subid, subid, ticks)


def run_benchmark(name):
    print(name)
    seconds = timeit(
        '%s()' % name, setup='from __main__ import %s' % name, number=10000)
    print(seconds)
    print()


if __name__ == '__main__':
    for item_name in globals().keys():
        if not item_name.startswith('bench_'):
            continue
        run_benchmark(item_name)



#!/usr/bin/env python

import os

from setuptools import setup, find_packages

from version import get_git_version
VERSION, SOURCE_LABEL = get_git_version()
PROJECT = 'dossier.label'
AUTHOR = 'Diffeo, Inc.'
AUTHOR_EMAIL = 'support@diffeo.com'
URL = 'https://github.com/dossier/dossier.label'
DESC = 'Label (ground truth) storage for DossierStack'


def read_file(file_name):
    file_path = os.path.join(
        os.path.dirname(__file__),
        file_name
        )
    return open(file_path).read()


setup(
    name=PROJECT,
    version=VERSION,
    description=DESC,
    license='MIT',
    long_description=read_file('README.rst'),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    namespace_packages=['dossier'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
    ],
    install_requires=[
        'cbor',
        'enum34',
        'kvlayer>=0.5.7',
        'pyquchk',
        'pytest',
        'total-ordering',
        'yakonfig >= 0.7.2',
    ],
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'dossier.label = dossier.label.run:main',
        ],
    },
)



'''Unit tests for the ``dossier.label`` tool.

.. This software is released under an MIT/X11 open source license.
   Copyright 2012-2015 Diffeo, Inc.

'''
from __future__ import absolute_import
from cStringIO import StringIO

import cbor
import pytest

from kvlayer._local_memory import LocalStorage

from dossier.label import CorefValue, Label, LabelStore
from dossier.label.run import App


@pytest.fixture
def kvlclient():
    c = LocalStorage(app_name='a', namespace='n')
    c._data = {}
    return c


@pytest.fixture
def label_store(kvlclient):
    return LabelStore(kvlclient)


@pytest.fixture
def app(label_store):
    a = App()
    a._label_store = label_store
    a.stdout = StringIO()
    return a


@pytest.fixture
def four_labels(label_store):
    labels = [
        Label('c1', 'c2', 'a1', CorefValue.Positive, epoch_ticks=1234567890),
        Label('c1', 'c2', 'a1', CorefValue.Positive, epoch_ticks=1234567891),
        Label('c1', 'c2', 'a1', CorefValue.Positive, epoch_ticks=1234567892),
        Label('c1', 'c2', 'a2', CorefValue.Negative, epoch_ticks=1234567890),
    ]
    for label in labels:
        label_store.put(label)
    return labels


@pytest.fixture
def four_item_dump():
    return [
        {u'content_id1': 'c1',
         u'content_id2': 'c2',
         u'annotator_id': 'a1',
         u'value': 1,
         u'subtopic_id1': '',
         u'subtopic_id2': '',
         u'epoch_ticks': 1234567892,
         u'rating': 1},
        {u'content_id1': 'c1',
         u'content_id2': 'c2',
         u'annotator_id': 'a1',
         u'value': 1,
         u'subtopic_id1': '',
         u'subtopic_id2': '',
         u'epoch_ticks': 1234567891,
         u'rating': 1},
        {u'content_id1': 'c1',
         u'content_id2': 'c2',
         u'annotator_id': 'a1',
         u'value': 1,
         u'subtopic_id1': '',
         u'subtopic_id2': '',
         u'epoch_ticks': 1234567890,
         u'rating': 1},
        {u'content_id1': 'c1',
         u'content_id2': 'c2',
         u'annotator_id': 'a2',
         u'value': -1,
         u'subtopic_id1': '',
         u'subtopic_id2': '',
         u'epoch_ticks': 1234567890,
         u'rating': 0},
    ]


@pytest.fixture
def two_item_dump(four_item_dump):
    return [four_item_dump[0], four_item_dump[3]]


def test_list_short(app, label_store):
    label_store.put(Label('c1', 'c2', 'annotator', CorefValue.Positive,
                          epoch_ticks=1234567890))

    app.runcmd('list', [])

    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by annotator at 2009-02-13 23:31:30\n')


def test_list_two(app, label_store):
    label_store.put(Label('c1', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890))
    label_store.put(Label('c1', 'c2', 'a2', CorefValue.Negative,
                          epoch_ticks=1234567890))

    app.runcmd('list', [])

    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:30\n'
            'c1 !=(0) c2 by a2 at 2009-02-13 23:31:30\n')


def test_list_collapses(app, four_labels):
    app.runcmd('list', [])
    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:32\n'
            'c1 !=(0) c2 by a2 at 2009-02-13 23:31:30\n')


def test_list_deleted(app, four_labels):
    app.runcmd('list', ['--include-deleted'])
    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:32\n'
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:31\n'
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:30\n'
            'c1 !=(0) c2 by a2 at 2009-02-13 23:31:30\n')


def test_list_subtopics(app, label_store):
    label_store.put(Label('c1', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890,
                          subtopic_id1='s1', subtopic_id2='s2'))

    app.runcmd('list', [])

    assert (app.stdout.getvalue() ==
            'c1(s1) ==(1) c2(s2) by a1 at 2009-02-13 23:31:30\n')


def test_dump_all(app, four_labels, four_item_dump):
    app.runcmd('dump_all', [])
    dump = cbor.loads(app.stdout.getvalue())
    assert dump == four_item_dump


def test_dump_all_exclude_deleted(app, four_labels, two_item_dump):
    app.runcmd('dump_all', ['--exclude-deleted'])
    dump = cbor.loads(app.stdout.getvalue())
    assert dump == two_item_dump


def test_load_short(app, label_store, two_item_dump, tmpdir):
    fn = tmpdir.join('dump.cbor')
    fn.write(cbor.dumps(two_item_dump), mode='wb')

    app.runcmd('load', [str(fn)])
    assert (len(list(label_store.everything(include_deleted=True))) ==
            len(two_item_dump))
    assert (len(list(label_store.everything(include_deleted=False))) ==
            len(two_item_dump))
    assert app.stdout.getvalue() == ''
    app.runcmd('dump_all', [])
    assert cbor.loads(app.stdout.getvalue()) == two_item_dump


def test_load_full(app, label_store, four_item_dump, tmpdir):
    fn = tmpdir.join('dump.cbor')
    fn.write(cbor.dumps(four_item_dump), mode='wb')

    app.runcmd('load', [str(fn)])
    assert (len(list(label_store.everything(include_deleted=True))) ==
            len(four_item_dump))
    assert (len(list(label_store.everything(include_deleted=False))) == 2)
    assert app.stdout.getvalue() == ''
    app.runcmd('dump_all', [])
    assert cbor.loads(app.stdout.getvalue()) == four_item_dump


def test_get(app, four_labels):
    app.runcmd('get', ['c1'])
    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:32\n'
            'c1 !=(0) c2 by a2 at 2009-02-13 23:31:30\n')


def test_get_missing(app, four_labels):
    app.runcmd('get', ['c0'])
    assert app.stdout.getvalue() == ''


def test_get_negative(app, four_labels):
    app.runcmd('get', ['--value', '-1', 'c1'])
    assert (app.stdout.getvalue() ==
            'c1 !=(0) c2 by a2 at 2009-02-13 23:31:30\n')


def test_connected(app, label_store):
    label_store.put(Label('c1', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890))
    label_store.put(Label('c3', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890))

    app.runcmd('connected', ['c1'])
    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:30\n'
            'c2 ==(1) c3 by a1 at 2009-02-13 23:31:30\n')


def test_connected_two(app, label_store):
    label_store.put(Label('c1', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890))
    label_store.put(Label('c3', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890))

    app.runcmd('connected', ['c2'])
    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:30\n'
            'c2 ==(1) c3 by a1 at 2009-02-13 23:31:30\n')


def test_connected_negative(app, label_store):
    label_store.put(Label('c1', 'c2', 'a1', CorefValue.Positive,
                          epoch_ticks=1234567890))
    label_store.put(Label('c3', 'c2', 'a1', CorefValue.Negative,
                          epoch_ticks=1234567890))

    app.runcmd('connected', ['c2'])
    assert (app.stdout.getvalue() ==
            'c1 ==(1) c2 by a1 at 2009-02-13 23:31:30\n')
